
#ifndef auto_rfutils_h
#define auto_rfutils_h 1

#include "AutoRandomFieldsUtilsLocal.h"

#define MAXUNITS 4
#define MAXCHAR 18 // max number of characters for (covariance) names  
#define RFOPTIONS "RFoptions"
#define CLASS_TRYERROR "try-error"

#define WARN_UNKNOWN_OPTION_ALL 4
#define WARN_UNKNOWN_OPTION_SINGLE 3
#define WARN_UNKNOWN_OPTION_CAPITAL 2
#define WARN_UNKNOWN_OPTION_NONE1 1
#define WARN_UNKNOWN_OPTION_NONE 0

#define CONTACT " Please contact the maintainer martin.schlather@math.uni-mannheim.de.\n"

#endif
